unit MDServerUnit;

interface

uses
  Windows, ActiveX, Classes, ComObj, ExMegaDemoServer_TLB, StdVcl, Registry,
  SysUtils, SxTools, Controls;

type
  TMDServer = class(TTypedComObject, IMDServer)
  protected
    function GetStgFileName(out Value: WideString): HResult; stdcall;
    function SetStgFileName(const FileName: WideString): HResult; stdcall;
    function CreateStorage(const FileName: WideString): HResult; stdcall;
    function GetFileStorage(out FileStg: IMDFileObj): HResult; stdcall;
    function GetRelativeFileStorage(const pPath: WideString;
      out pFileStg: IMDFileObj): HResult; stdcall;
    function DeteleFileStgElement(const ElPath: WideString): HResult; stdcall;
    function TestServer(Value: Integer): HResult; stdcall;
    {Declare IMDServer methods here}
  public
    procedure Initialize;override;
  end;

var
  MDStorage:IStorage;
  FileStorage:IStorage;
  StorageFileName:WideString;
  FileStorageObject:IMDFileObj;
  dwRegister:Integer;

procedure CloseStorage;

implementation

uses ComServ, FileObjUnit, Dialogs;

procedure CloseStorage;
begin
  RevokeActiveObject(dwRegister,nil);
  FileStorageObject:=nil;
  FileStorage:=nil;
  MDStorage:=nil;
  StorageFileName:='';
end;

procedure OpenStorage;
var
  Reg:TRegistry;
  tempFileName:WideString;
begin

  if ComServer.StartMode in [smRegServer, smUnregServer] then exit;

  CloseStorage;

  Reg:=TRegistry.Create;
  try
    Reg.OpenKey('\Software\ShellPlus\Examples\MegaDemoNSE',True);
    if Reg.ValueExists('StorageFile') then
       tempFileName:=Reg.ReadString('StorageFile');
  finally
    FreeAndNil(Reg);
  end;
  if not FileExists(tempFileName) then
  begin
    // Opening properties of Root folder
    SxExecutePropertiesDialog('::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\::{24AF2956-9368-4655-B977-8D4EC810A219}');
    Exit;
  end;
  if Failed(StgOpenStorage(PWideChar(tempFileName),nil,STGM_READWRITE or
                                             STGM_SHARE_EXCLUSIVE or
                                             STGM_DIRECT,nil,0,MDStorage)) then
     MDStorage:=nil
  else
     StorageFileName:=tempFileName;
end;

function TMDServer.GetStgFileName(out Value: WideString): HResult;
begin
  Value:=StorageFileName;
  Result:=S_OK;
end;

function TMDServer.SetStgFileName(const FileName: WideString): HResult;
var
  Reg:TRegistry;
begin
  try
    Reg:=TRegistry.Create;
    try
      Reg.OpenKey('\Software\ShellPlus\Examples\MegaDemoNSE',True);
      Reg.WriteString('StorageFile',FileName);
    finally
      FreeAndNil(Reg);
    end;
    OpenStorage;
    Result:=S_OK;
  except
    Result:=E_FAIL;
  end;
end;

function TMDServer.CreateStorage(const FileName: WideString): HResult;
var
  tempStorage:IStorage;
begin
  Result:=S_OK;
  try
    if FileExists(FileName) then
       DeleteFile(FileName);
    if Succeeded(StgCreateDocFile(PWideChar(FileName),STGM_READWRITE or
                                             STGM_SHARE_EXCLUSIVE or
                                             STGM_DIRECT,0,tempStorage)) then
    begin
      tempStorage:=nil;
      SetStgFileName(FileName);
    end
    else
    begin
      MDStorage:=nil;
      StorageFileName:='';
    end;
  except
    Result:=E_FAIL;
  end;
end;

function TMDServer.GetFileStorage(out FileStg: IMDFileObj): HResult;
var
  tempFileStorage:IStorage;
  tempFileStg:IFileObjInit;
begin
  Result:=E_FAIL;
  if FileStorageObject=nil then
  begin
    if MDStorage=nil then
    begin
      OpenStorage; // trying to open storage once again
      if MDStorage=nil then
         Exit;
    end;
    if FileStorage=nil then
    begin
      Result:=MDStorage.OpenStorage(nameFileStorage,nil,stgMode,nil,0,tempFileStorage);
      if Failed(Result) then
         Result:=MDStorage.CreateStorage(nameFileStorage,stgMode or STGM_CREATE,0,0,tempFileStorage);
      if Failed(Result) then
         Exit;
      FileStorage:=tempFileStorage;
      MDStorage.Commit(STGC_DEFAULT);
    end;
    FileStg:=CoMDFileObj.Create;
    tempFileStg:=FileStg as IFileObjInit;
    tempFileStg.SetStorage(FileStorage, nil);
    FileStorageObject:=FileStg;
  end
  else
    FileStg:=FileStorageObject;
  Result:=S_OK;
end;

function TMDServer.GetRelativeFileStorage(const pPath: WideString;
  out pFileStg: IMDFileObj): HResult;
var
  tempStr:TStrings;
  i:Integer;
  tempParentStg:IMDFileObj;
begin // will receive paths, that only relative to root
  Result:=E_FAIL;
  pFileStg:=nil;
  tempStr:=TStringList.Create;
  try
    FilePathToStrings(pPath,tempStr);
    if tempStr.Count=0 then
       Exit;
    GetFileStorage(pFileStg);
    if pFileStg=nil then
       Exit;
    tempStr.Delete(0); // this will always be a {A0283B15-5B9C-4435-9056-33881A4663D0} value
    if tempStr.Count=0 then
    begin // this was a root folder
      Result:=S_OK;
      Exit;
    end;
    for i:=0 to tempStr.Count-1 do
    begin
      tempParentStg:=pFileStg;
      pFileStg:=nil;
      tempParentStg.GetChildByName(tempStr[i],pFileStg);
      if pFileStg=nil then
         Break;
    end;
    if pFileStg<>nil then
       Result:=S_OK;
  finally
    FreeAndNil(tempStr);
  end;
end;

function TMDServer.DeteleFileStgElement(const ElPath: WideString): HResult;
var
  tempStrl:TStrings;
  tempFileObj:IMDFileObj;
  eltName:String;
begin
  Result:=E_FAIL;
  tempStrl:=TStringList.Create;
  try
    FilePathToStrings(ElPath,tempStrl);
    if tempStrl.Count<=1 then
    begin
      Result:=S_FALSE;
      Exit;
    end;
    GetRelativeFileStorage(ElPath,tempFileObj);
    if (tempFileObj=nil) or (tempFileObj.ReadOnly) then
    begin
      Result:=E_ACCESSDENIED;
      Exit;
    end;
    eltName:=tempStrl[tempStrl.Count-1];
    tempStrl.Delete(tempStrl.Count-1);
    GetRelativeFileStorage(StringsToFilePath(tempStrl),tempFileObj);
    if tempFileObj=nil then
       Exit;
    tempFileObj.DeleteChild(eltName);
    Result:=S_OK;
  finally
    FreeAndNil(tempStrl);
  end;
end;

procedure TMDServer.Initialize;
begin
  inherited;
  RegisterActiveObject(Self,CLASS_MDServer,ACTIVEOBJECT_STRONG,dwRegister);
end;

function TMDServer.TestServer(Value: Integer): HResult;
begin
  Result:=Value;
end;

initialization
  TTypedComObjectFactory.Create(ComServer, TMDServer, Class_MDServer, ciMultiInstance, tmApartment);

  OpenStorage;

finalization
  CloseStorage;
end.
